from enstools.feature.util.data_utils import pb_str_to_datetime

class DataGraph:
    """
    Abstraction for graphs:
    A tracked set (graph in protobuf desc populated) is needed to create this object.
    It contains methods to iterate a graph and to generate the tracks.
    """

    def __init__(self, set_desc, tracking_tech):
        self.set_desc = set_desc
        self.graph = set_desc.graph
        self.tr_tech = tracking_tech

    def get_parents(self, node):
        """
        Get the parents of a node (previous timestep)
        Parameters
        ----------
        node: The node to get the parents of

        Returns
        -------
        List of nodes representing the parents. Multiple nodes indicate a merge event.
        """
        parents = []
        for edge in self.graph.edges:
            for child in edge.childs:
                if child == node:
                    parents.append(edge.parent)
        return parents

    def get_childs(self, node):
        """
        Get the childs of a node (next timestep)
        Parameters
        ----------
        node: The node to get the childs of

        Returns
        -------
        List of nodes representing the childs. Multiple nodes indicate a split event.
        """
        for edge in self.graph.edges:
            if edge.parent == node:
                return edge.childs
        return []

    def generate_tracks(self, apply_filter=True):
        """
        After tracking graph has been computed, here, tracks can be computed, which are a disjoint subset of graphs
        of the total graph. It is based on a simple heuristic. The nodes are ordered by time. For each non-classified
        node, all downstream nodes are searched. If any of these nodes is already classified, use the same ID.
        Otherwise give this stream a new id (new track).

        Returns
        -------
        Tracks are added to the given set inplace, also a list of Tracks is returned (list of Graphs)
        """

        # sort nodes by time of first node (is in key), as list here
        time_sorted_nodes = list(sorted(self.graph.edges, key=lambda item: item.parent.time))

        id_per_node = [None] * len(time_sorted_nodes)

        cur_id = 0

        # iterate over all time sorted identified connections
        # search temporal downstream tracked troughs and group them using a set id
        for con_idx, oc in enumerate(time_sorted_nodes):

            if id_per_node[con_idx] is not None:  # already part of an object
                continue

            # not part of object -> get (temporal) downstream connections = part of same object (return indices of them)
            downstream_node_indices = DataGraph.get_downstream_node_indices(time_sorted_nodes, con_idx)

            # any of downstream nodes already part of an object?
            connected_id = None
            for ds_node_idx in downstream_node_indices:
                if id_per_node[ds_node_idx] is not None:
                    # if connected_object_id is not None:
                    #     print("Double ID, better resolve") # TODO
                    connected_id = id_per_node[ds_node_idx]

            # if so set all nodes to this found id
            if connected_id is not None:
                for ds_node_idx in downstream_node_indices:
                    id_per_node[ds_node_idx] = connected_id
                continue

            # else new path for all object_nodes
            cur_id_needs_update = False
            for ds_node_idx in downstream_node_indices:
                id_per_node[ds_node_idx] = cur_id
                cur_id_needs_update = True

            if cur_id_needs_update:
                cur_id += 1

        # done, now extract every object by id and put them into subgraphs
        tracks = []
        for object_id in range(cur_id):
            object_idxs = [i for i in range(len(id_per_node)) if id_per_node[i] == object_id]

            cur_object_nodes = [time_sorted_nodes[i] for i in object_idxs]  # states of this object
            # cur_troughs = cur_troughs.sortbytime # already sorted?

            track = self.tr_tech.pb_reference.ObjectGraph()
            track.edges.extend(cur_object_nodes)

            if apply_filter:
                # if apply_filter is set ask tracking technique if this track should be kept.
                if not self.tr_tech.keep_track(track):
                    continue  # dont keep

            # add to desc
            self.set_desc.tracks.append(track)

            # create return object
            track_set = self.tr_tech.pb_reference.TrackableSet()
            track_set.CopyFrom(self.set_desc)
            track_set.graph.CopyFrom(track)
            del track_set.tracks[:]

            tracks.append(DataGraph(track_set, self.tr_tech))

        print("Generated tracks for current set: " + str(cur_id) + " -> Filtered: " + str(len(tracks)))
        return tracks

    @staticmethod
    def get_downstream_node_indices(graph_list, start_idx, until_time=None):
        """
        Helper method for the track generation. Searches all downstream node indices.

        Returns
        -------
        list of downstream indices in list
        """
        node_indices = [start_idx]
        obj_node_list = [con_.parent for con_ in graph_list]  # this up

        co = graph_list[start_idx]
        node, connected_nodes = co.parent, co.childs

        for c_node in connected_nodes:
            # if until_time is set: if c_node time later than until_time end recursion here
            if until_time is not None and pb_str_to_datetime(c_node.time) > until_time:
                continue

            # get index of connected node in graph
            c_node_idx = obj_node_list.index(c_node)

            # call recursively on this connected node
            c_node_downstream_indices = DataGraph.get_downstream_node_indices(graph_list, c_node_idx,
                                                                              until_time=until_time)
            node_indices.extend(c_node_downstream_indices)

        return list(set(node_indices))
